import typing
import inflection
import digi


# Each digi can be a data source that has egress attributes.
# Each egress attribute maps to a data pool branch, e.g., l1@main.
# A source is uniquely identified by its group, version, kind,
# name, namespace, and the egress ID/name. The source attribute
# is parsed with the following rules:
# - If all attributes are given, a single egress branch is identified.
# - If the egress ID is omitted, default to main branch at the source pool.
# - If the name is omitted, default to all sources with of same kind mounted
#   to the destination.
# - If the group / version / kind is omitted, default to the destination's
#   corresponding attribute.
#   - XXX self-reference is allowed with ingress pointing to the
#     destination itself
# TBD add tests
def parse_source(source: typing.Union[dict, str], *,
                 exist_only: bool = True) -> typing.List[str]:
    """
    Return the list of egress pool@branch given source attributes.
    """
    if isinstance(source, dict):
        branch = source.get("egress", "main")
        # XXX assume name is a unique id to digi and ignore namespace
        if "name" in source:
            return [f"{source['name']}@{branch}"]
        else:
            group = source.get("group", digi.group)
            version = source.get("version", digi.version)
            kind = source.get("kind", digi.kind)
    elif isinstance(source, str):
        parts = source.split("@")
        branch = parts[1] if len(parts) > 1 else "main"
        if "kind:" not in parts[0]:
            return [f"{parts[0]}@{branch}"]
        else:
            # kind:g/v/k@main or kind:k@main; k <-> r
            kind = parts[0].lstrip("kind:")
            group, version, kind = digi.util.parse_kind(kind)
    else:
        raise NotImplementedError

    if kind == "any":
        mounts = digi.model.get_mount(any=True)
    else:
        mounts = digi.model.get_mount(group, version, inflection.pluralize(kind))
    pools = [digi.util.trim_default_namespace(name) for name in mounts.keys()] \
        if mounts else []

    return [f"{pool}@{branch}"
            for pool in pools if not exist_only
            or digi.data.lake.branch_exist(pool, branch)]



import sys
import json
import threading
import datetime
from abc import ABC, abstractmethod
from typing import List, Callable

import digi
from digi.data import logger, zjson
from digi.data import zed, sync
from digi.data import router


class Pool(ABC):
    @abstractmethod
    def __init__(self, name: str):
        self.name = name
        self.lock = threading.Lock()

    @abstractmethod
    def load(self, objects: List[dict]):
        raise NotImplementedError

    @abstractmethod
    def query(self, query: str):
        raise NotImplementedError

    @abstractmethod
    def watch(self, once: Callable, *,
              branch: str = "main"):
        raise NotImplementedError

    @abstractmethod
    def create_branch_if_not_exist(self, query: str):
        raise NotImplementedError


class ZedPool(Pool):
    def __init__(self, name):
        super().__init__(name)
        self.client = zed.Client(
            base_url=digi.data.lake_url
        )

    def load(self, objects: List[dict], *,
             branch="main",
             encoding="zjson",
             same_type=False):
        # update event and processing time
        now = datetime.datetime.now()
        if encoding == "zjson":
            for o in objects:
                # event_ts will be attached at the first
                # data router if does exist from the source
                if "event_ts" not in o:
                    o["event_ts"] = o.get("ts", now)
                o["ts"] = now
            data = "".join(zjson.encode(objects))
        elif encoding == "json":
            zjson_now = zjson.encode_datetime(now)  # as str
            for o in objects:
                if "event_ts" not in o:
                    o["event_ts"] = o.get("ts", zjson_now)
                o["ts"] = zjson_now
            data = "\n".join(json.dumps(o) for o in objects)
        else:
            raise NotImplementedError

        self.lock.acquire()
        try:
            # TBD better source name
            meta = json.dumps({f"{digi.name}": zjson.encode_datetime(now)})
            self.client.load(self.name, data,
                             branch_name=branch,
                             commit_author=digi.name,
                             meta=meta)
            # TBD load from digi also commits source ts in meta
        except Exception as e:
            digi.logger.warning(f"unable to load "
                                f"{data} to {self.name}: {e}")
        finally:
            self.lock.release()

    def query(self, query: str):
        if query != "":
            query = f"| {query}"
        query = f"from {self.name} {query}"
        return self.client.query(query)

    def watch(self, fn: Callable, *,
              in_flow: str = "",
              eoio: bool = True,
              ):
        """Watch changes of the main pool and run UDF."""
        source = f"{self.name}@main"
        return sync.Watch(fn,
                          sources=[source],
                          eoio=eoio,
                          in_flow=in_flow)

    def create_branch_if_not_exist(self, branch: str):
        if not self.client.branch_exist(self.name, branch):
            self.client.create_branch(self.name, branch)
            self.load([router.Egress.INIT], branch=branch)


def pool_name(g, v, r, n, ns):
    _, _, _ = g, v, r
    if ns == "default":
        return f"{n}"
    else:
        # TBD update digi creation
        return f"{ns}_{n}"


providers = {
    "zed": ZedPool
    # ...
}


def create_pool():
    global providers
    if digi.pool_provider == "":
        digi.pool_provider = "zed"

    if digi.pool_provider in {"none", "false"}:
        return None

    if digi.pool_provider not in providers:
        logger.fatal(f"unknown pool provider {digi.pool_provider}")
        sys.exit(1)

    return providers[digi.pool_provider](
        pool_name(*digi.duri)
    )



import json
import getpass
import urllib
import zed
from . import zjson


class Client(zed.Client):
    """TBD patch upstream"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(self, pool_name_or_id, data, branch_name='main',
             commit_author=getpass.getuser(), commit_body='', meta=''):
        pool = urllib.parse.quote(pool_name_or_id)
        branch = urllib.parse.quote(branch_name)
        url = self.base_url + '/pool/' + pool + '/branch/' + branch
        commit_message = {'author': commit_author, 'body': commit_body, 'meta': meta}
        headers = {'Zed-Commit': json.dumps(commit_message)}
        r = self.session.post(url, headers=headers, data=data)
        self.__raise_for_status(r)

    def create_branch(self, pool, name, *,
                      commit=f"0x{'0' * 40}"):
        r = self.session.post(self.base_url + f"/pool/{pool}",
                              json={
                                  "name": name,
                                  "commit": commit,
                              })
        self.__raise_for_status(r)

    def branch_exist(self, pool, name):
        records = self.query(f"from {pool}:branches")
        return name in set(r["branch"]["name"] for r in records)

    def query(self, query):
        return zjson.decode_raw(self.query_raw(query))



import copy
import typing
import traceback
from collections import OrderedDict

import digi
import digi.util as util
import digi.filter as filter_
import digi.processor as processor


class HandlerType:
    BUILTIN = 1
    REFLEX = 2


class __Reconciler:
    def __init__(self):
        # handlers are stored as tuples (fn, condition, path, priority)
        # the higher the priority value, the higher the priority;
        # default priority is 0; low priority handlers are run first;
        # priority lower than 0 is skipped.
        # - condition: a function that decides whether the handler
        #   should be run or not.
        # - path: the attribute subtree the handler subscribes to

        # sorted list of handlers in execution order
        self.handlers = list()
        self.g = digi.g
        self.v = digi.v
        self.r = digi.r
        self.n = digi.n
        self.ns = digi.ns
        self.auri = digi.auri

        self._logger = digi.logger

        self.skip_gen = -1
        self._skip = True
        self.last_seen_gen = -1
        self.count = 0

        # handler info (e.g., priority) are used to
        # generate the self.handlers upon handler updates;
        # each of form (fn, condition, view_path, priority, type)
        # keyed by handler's name (fn.__name__)
        self._handler_info = OrderedDict()
        self._handler_info_updated = True
        # if a handler's update didn't get applied to the model
        # the handler is marked as pending and will be invoked
        # in future reconciliation to be executed again. The
        # handler object's id is used as the key.
        self._pending_handler = set()

        # most recent view of the model, served as an in-memory
        # read-only copy to be used in external application
        self._view = dict()

        self._data_watches = dict()

    def run(self, spec, old, diff, *args, **kwargs):
        spec = dict(spec)
        proc_spec = dict(spec)

        self._view = proc_spec
        self._update_handler_info(spec, diff)
        self._compile_handler()

        for fn, cond, path, _ in self.handlers:
            if cond(proc_spec, diff, path, *args, **kwargs) \
                    or id(fn) in self._pending_handler:
                # handler edits the spec object
                try:
                    # TBD allow subview to be a forest
                    fn(
                        subview=safe_lookup(proc_spec, path),
                        proc_view=proc_spec,
                        view=spec, old_view=old,
                        mount=proc_spec.get("mount", {}),
                        obs=proc_spec.get("obs", {}),
                        back_prop=get_back_prop(diff),
                        diff=diff,
                        meta=proc_spec.get("meta", {}),
                    )
                    self._pending_handler.add(id(fn))
                except Exception as e:
                    self._logger.error(f"reconcile error: {e}")
                    self._logger.error(traceback.format_exc())
                    # TBD: expose driver status on model, e.g., obs.reason/or some debug attribute
                    return proc_spec
        return proc_spec

    def add(self, handler: typing.Callable,
            condition: typing.Callable,
            priority: int,
            path: tuple = (),
            typ=HandlerType.BUILTIN):

        n = handler.__name__

        if n in self._handler_info:
            # TBD perhaps use deterministic name
            n = n + util.uuid_str()

        self._handler_info[n] = {
            "fn": handler,
            "condition": condition,
            "view_path": path,
            "priority": priority,
            "type": typ,
        }

    def _update_handler_info(self, spec, diff):
        # check whether there is a reflex change
        _changed = not all(len(_path) < 2 or _path[1] != "reflex"
                           for _, _path, _, _ in diff)

        if _changed:
            reflexes = util.deep_get(spec, "reflex", {})

            # trim reflexes
            to_remove = list()
            for n, info in self._handler_info.items():
                if info["type"] == HandlerType.BUILTIN:
                    continue
                if n not in reflexes:
                    to_remove.append(n)

            for n in to_remove:
                self._handler_info.pop(n, {})

            # update handlers
            for n, r in reflexes.items():
                info = self._handler_info.get(n, {
                    "fn": do_nothing,
                    "condition": filter_.always,  # TBD conditioned reflex
                    "view_path": ".",
                    "priority": 0,
                    "type": HandlerType.REFLEX,
                })

                patch = dict()
                if "policy" in r:
                    patch.update({
                        "fn": self._new_reflex(r["policy"],
                                               r.get("processor", "py"))
                    })

                if "priority" in r:
                    patch.update({
                        "priority": r["priority"]
                    })

                info.update(patch)
                self._handler_info[n] = info

            self._handler_info_updated = True

    @staticmethod
    def _new_reflex(logic, proc="py"):
        if logic is None:
            return do_nothing
        if proc == "py":
            return processor.py(logic)
        if proc == "jq":
            return processor.jq(logic)
        ...

    def _compile_handler(self):
        if not self._handler_info_updated:
            return

        self.handlers = list()
        for _, hi in self._handler_info.items():
            # treat negative priority as disabled
            if hi["priority"] < 0:
                continue
            self.handlers.append((hi["fn"], hi["condition"],
                                  hi["view_path"], hi["priority"]))

        # sort by priority
        self.handlers = sorted(self.handlers, key=lambda x: x[3])
        self._handler_info_updated = False

    def view(self):
        return copy.deepcopy(self._view)

    def clear_pending(self):
        self._pending_handler.clear()

    def add_data_watch(self, name, watch):
        self._data_watches[name] = watch

    def start_data_watch(self):
        for name, watch in self._data_watches.items():
            # TBD check if this egress is on the model, if not
            #  do digi.model.patch()
            # TBD we assume egress name is the same as watch's
            digi.pool.create_branch_if_not_exist(name)
            watch.start()
            digi.logger.info(f"data watch {name} started")

    def do_not_skip(self):
        """Set to ensure the next reconciliation loop won't be
        skipped even if the last generation is done by the digi.
        The flag is reset upon the should_skip() call."""
        self._skip = False

    def should_skip(self):
        skip = self._skip
        self._skip = True
        return skip


def safe_lookup(d: dict, path: tuple):
    if path == (".",):
        return d

    for k in path:
        d = d.get(k, {})
    return d


def do_nothing(*args, **kwargs):
    _, _ = args, kwargs


def get_back_prop(diff):
    bp = list()
    for op, path, old, new in diff:
        if op != "change" and op != "add":
            continue
        if len(path) < 3 or path[0] != "spec" \
                or path[1] != "mount":
            continue

        fs = set(path)
        if "intent" not in fs and "input" not in fs:
            continue
        bp.append((op, path, old, new))
    return bp


rc = __Reconciler()



"""Predefined dataflows."""

refresh_ts = """switch ( 
    case has(event_ts) => yield this | put ts := now()
    case has(ts) => put event_ts := ts | put ts := now() 
    default => put event_ts := now() | put ts := event_ts
)"""

patch_ts = "switch ( case has(ts) => yield this " \
           "default => put ts := now() )"

drop_meta = "not __meta"



import os
import uuid
import time
import asyncio
import contextlib
import threading

import digi
import inflection
import logging
from typing import (
    Tuple, Callable, Union, Any, Iterable
)
from functools import reduce

import kubernetes
from kubernetes import config
from kubernetes.client.rest import ApiException

import kopf
from kopf._core.intents.registries import SmartOperatorRegistry as KopfRegistry

logger = logging.getLogger(__name__)
logger.setLevel(digi.log_level)

try:
    # use service config
    config.load_incluster_config()
except:
    # use kubeconfig
    config.load_kube_config()


class DriverError:
    GEN_OUTDATED = 41


_api = kubernetes.client.CustomObjectsApi()


def run_operator(registry: KopfRegistry,
                 log_level=logging.INFO,
                 skip_log_setup=False,
                 ) -> (threading.Event, threading.Event):
    clusterwide = os.environ.get("CLUSTERWIDE", True)
    kopf_logging = os.environ.get("KOPFLOG", "true") == "true"
    if not kopf_logging:
        kopf_logger = logging.getLogger("kopf")
        kopf_logger.propagate = False
        kopf_logger.handlers[:] = [logging.NullHandler()]

    ready_flag = threading.Event()
    stop_flag = threading.Event()

    def kopf_thread():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        with contextlib.closing(loop):
            if not skip_log_setup:
                # XXX (@kopf) disallow kopf configure root logger
                kopf.configure(verbose=log_level <= logging.DEBUG,
                               debug=log_level <= logging.DEBUG,
                               quiet=kopf_logging and log_level <= logging.INFO)
            loop.run_until_complete(kopf.operator(
                ready_flag=ready_flag,
                stop_flag=stop_flag,
                registry=registry,
                clusterwide=clusterwide,
            ))

    thread = threading.Thread(target=kopf_thread)
    thread.start()
    logger.info(f"started an operator")
    return ready_flag, stop_flag


class NamespacedName:
    def __init__(self, n, ns="default"):
        self.name = n
        self.namespace = ns


def uuid_str(length=4):
    return str(uuid.uuid4())[:length]


def spaced_name(n, ns) -> str:
    return f"{ns}/{n}"


def trim_default_space(n):
    return n.replace("default/", "")


trim_namespace = trim_default_namespace = trim_default_space


def simple_name(n):
    return trim_default_space(n)


def trim_gv(gvr):
    return gvr.split("/")[-1]


def gvr_from_body(b):
    g, v = tuple(b["apiVersion"].split("/"))
    r = inflection.pluralize(b["kind"].lower())
    return g, v, r


def trim_attr(spec: dict, attrs: set) -> dict:
    # BFS
    to_visit = [spec]
    for n in to_visit:
        to_trim = list()
        if type(n) is not dict:
            continue
        for k, v in n.items():
            if k not in attrs:
                to_visit.append(v)
            else:
                to_trim.append(k)
        for k in to_trim:
            n.pop(k, {})
    return spec


def apply_diff(model: dict, diff: list) -> dict:
    for op, fs, old, new in diff:
        if len(fs) == 0:
            continue
        if op in {"add", "change"}:
            n = model
            for f in fs[:-1]:
                if f not in n:
                    n[f] = dict()
                n = n[f]
            n[fs[-1]] = new
    return model


def parse_spaced_name(nsn) -> Tuple[str, str]:
    parsed = nsn.split("/")
    if len(parsed) < 2:
        return parsed[0], "default"
    # name, namespace
    return parsed[1], parsed[0]


def parse_gvr(gvr: str) -> Tuple[str, str, str]:
    parsed = tuple(gvr.lstrip("/").split("/"))
    if len(parsed) == 1:
        return digi.group, digi.version, parsed[0]
    elif len(parsed) == 3:
        g, v, r = parsed
        return g, v, r
    else:
        raise Exception(f"{gvr} not in form of "
                        f"'[/]group/version/(kind or resource)'")


parse_kind = parse_gvr


def model_id(g, v, r, n, ns) -> str:
    return f"{g}/{v}/{r}/{spaced_name(n, ns)}"


def gvr(g, v, r) -> str:
    return f"{g}/{v}/{r}"


def gvr_from_kind(kind, g=digi.g, v=digi.v):
    return f"{g}/{v}/{inflection.pluralize(kind)}"


gvr_of = gvr_from_kind


def is_gvr(s: str) -> bool:
    return len(s.split("/")) == 3


def normalized_gvr(s, g, v) -> str:
    r = s.split("/")[-1]
    return f"{g}/{v}/{r}"


def normalized_nsn(s: str) -> str:
    if "/" not in s:
        return "default/" + s
    else:
        return s


def safe_attr(s):
    return s.replace(".", "-")


def parse_model_id(s) -> Tuple[str, str, str, str, str]:
    ps = s.lstrip("/").split("/")
    assert len(ps) in {4, 5}
    if len(ps) == 4:
        return ps[0], ps[1], ps[2], ps[3], "default"
    return ps[0], ps[1], ps[2], ps[4], ps[3]


def get_spec(g, v, r, n, ns) -> (dict, str, int):
    global _api

    try:
        o = _api.get_namespaced_custom_object(group=g,
                                              version=v,
                                              namespace=ns,
                                              name=n,
                                              plural=r,
                                              )
    except ApiException as e:
        logger.warning(f"unable to get model {n}:", e)
        return None
    return o.get("spec", {}), \
           o["metadata"]["resourceVersion"], \
           o["metadata"]["generation"]


def patch_spec(g, v, r, n, ns, spec: dict, rv=None):
    global _api
    try:
        resp = _api.patch_namespaced_custom_object(group=g,
                                                   version=v,
                                                   namespace=ns,
                                                   name=n,
                                                   plural=r,
                                                   body={
                                                       "metadata": {} if rv is None else {
                                                           "resourceVersion": rv,
                                                       },
                                                       "spec": spec,
                                                   },
                                                   )
        return resp, None
    except ApiException as e:
        return None, e


def check_gen_and_patch_spec(g, v, r, n, ns, spec, gen):
    # patch the spec atomically if the current gen is
    # less than the given spec
    while True:
        _, rv, cur_gen = get_spec(g, v, r, n, ns)
        if gen < cur_gen:
            e = ApiException()
            e.status = DriverError.GEN_OUTDATED
            e.reason = f"generation outdated {gen} < {cur_gen}"
            return cur_gen, None, e

        resp, e = patch_spec(g, v, r, n, ns, spec, rv=rv)
        if e is None:
            return cur_gen, resp, None
        if e.status == 409:
            logger.info(f"unable to patch {n} due to conflict; retry")
        else:
            logger.warning(f"patch error {e}")
            return cur_gen, resp, e


# utils
def put(path, src, target, transform=lambda x: x):
    if not isinstance(target, dict):
        return

    ps = path.split(".")
    for p in ps[:-1]:
        if p not in target:
            return
        target = target[p]

    if not isinstance(src, dict):
        if src is None:
            target[ps[-1]] = None
        else:
            target[ps[-1]] = transform(src)
        return

    for p in ps[:-1]:
        if p not in src:
            return
        src = src[p]
    target[ps[-1]] = transform(src[ps[-1]])


def deep_get(d: dict, path: Union[str, Iterable], default=None) -> Any:
    """Can accept paths in form root.'digi.dev/test'.control.power"""
    if "'" in path or '"' in path:
        path = path.replace('"', "'")
        parsed = list()
        for i, seg in enumerate(path.split("'")):
            if seg == "":
                continue
            if i % 2 != 0:
                parsed.append(seg)
            else:
                parsed += seg.strip(".").split(".")
        path = parsed
    else:
        path = path.split(".") if isinstance(path, str) else path
    return reduce(lambda _d, key: _d.get(key, default)
    if isinstance(_d, dict) else default, path, d)


def deep_set(d: dict, path: Union[str, Iterable], val: Any, create=True):
    if not isinstance(d, dict):
        return
    if isinstance(path, str):
        keys = path.split(".")
    elif isinstance(d, Iterable):
        keys = path
    else:
        return
    for k in keys[:-1]:
        if k not in d:
            if create:
                d[k] = {}
            else:
                return
        d = d[k]
    d[keys[-1]] = val


def get_inst(d: dict, gvr_str) -> dict:
    inst = dict()
    for _n, body in d.get(gvr_str, {}).items():
        i = body.get("spec", None)
        if i is not None:
            inst[_n] = i
    return inst


def deep_set_all(ds, path: str, val):
    if isinstance(ds, list):
        for d in ds:
            deep_set(d, path, val)
    elif isinstance(ds, dict):
        for _, d in ds.items():
            deep_set(d, path, val)


def mount_size(mounts: dict, gvr_set: set = None,
               has_spec=False, cond: Callable = lambda x: True):
    count = 0
    for typ, models in mounts.items():
        if gvr_set is not None and typ not in gvr_set:
            continue

        for _, m in models.items():
            if not cond(m):
                continue
            if has_spec:
                count += 1 if "spec" in m else 0
            else:
                count += 1
    return count


def typ_attr_from_child_path(child_path):
    return child_path[2], child_path[-2]


def first_attr(attr, d: dict):
    if type(d) is not dict:
        return None
    if attr in d:
        return d[attr]
    for k, v in d.items():
        v = first_attr(attr, v)
        if v is not None:
            return v
    return None


def first_type(mounts):
    if type(mounts) is not dict or len(mounts) == 0:
        return None
    return list(mounts.keys())[0]


def full_gvr(a: str) -> str:
    # full gvr: group/version/plural
    if len(a.split("/")) == 1:
        return "/".join([os.environ["GROUP"],
                         os.environ["VERSION"],
                         a])
    else:
        return a


def gvr_equal(a: str, b: str) -> bool:
    return full_gvr(a) == full_gvr(b)


# TBD deprecate
class Auri:
    def __init__(self, **kwargs):
        self.group = str(kwargs["group"])
        self.version = str(kwargs["version"])
        self.kind = str(kwargs["kind"]).lower()
        self.resource = inflection.pluralize(self.kind)
        self.name = str(kwargs["name"])
        self.namespace = str(kwargs["namespace"])
        self.path = str(kwargs["path"])

    def gvr(self) -> tuple:
        return self.group, self.version, self.resource

    def gvk(self) -> tuple:
        return self.group, self.version, self.kind

    def auri(self):
        return self.group, self.version, self.resource, self.name, self.namespace

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"" \
               f"/{self.group}" \
               f"/{self.version}" \
               f"/{self.kind}" \
               f"/{self.namespace}" \
               f"/{self.name}{self.path}"


def parse_auri(s: str) -> Auri or None:
    ps = s.partition(".")
    main_segs = ps[0].lstrip("/").split("/")
    attr_path = "" if len(ps) < 2 else "." + ps[2]

    parsed = {
        2: lambda x: {
            "group": "digi.dev",
            "version": "v1",
            "kind": x[0],
            "name": x[1],
            "namespace": "default",
            "path": attr_path,
        },
        3: lambda x: {
            "group": "digi.dev",
            "version": "v1",
            "kind": x[0],
            "name": x[2],
            "namespace": x[1],
            "path": attr_path,
        },
        5: lambda x: {
            "group": x[0],
            "version": x[1],
            "kind": x[2],
            "name": x[3],
            "namespace": x[4],
            "path": attr_path,
        },
    }.get(len(main_segs), lambda: {})(main_segs)

    if parsed is None:
        return None
    return Auri(**parsed)


class Loop(threading.Thread):
    def __init__(self, loop_fn: callable,
                 loop_interval: float = 1, *,
                 loop_interval_fn: Callable = None,
                 ):
        threading.Thread.__init__(self)
        self.loop_fn = loop_fn
        self.loop_interval = loop_interval
        self.loop_interval_fn: Callable = loop_interval_fn
        self._stop_flag = threading.Event()

    def run(self):
        self._stop_flag.clear()
        while not self._stop_flag.is_set():
            self.loop_fn()
            if self.loop_interval_fn is None:
                i = self.loop_interval
            else:
                i = self.loop_interval_fn()
            time.sleep(i)

    def stop(self):
        self._stop_flag.set()

    def reset(self, loop_interval: float = None,
              loop_interval_fn: Callable = None) -> None:
        if loop_interval is not None:
            self.loop_interval = loop_interval
            self.loop_interval_fn = loop_interval_fn
        self.stop()


class Loader(Loop):
    def __init__(self, load_fn: callable,
                 load_interval: float = 1, *,
                 load_interval_fn: Callable = None,
                 ):
        super().__init__(loop_fn=load_fn,
                         loop_interval=load_interval,
                         loop_interval_fn=load_interval_fn)

    def reset(self, load_interval: float = None,
              load_interval_fn: Callable = None) -> None:
        super().reset(loop_interval=load_interval,
                      loop_interval_fn=load_interval_fn)


_report_loops = dict()


def report_loop(fn: Callable):
    global _report_loops
    _report_loops[fn.__name__] = Loader(load_fn=fn)

    @digi.on.meta
    def do_meta(meta):
        i = meta.get("report_interval", -1)
        _report_loops[fn.__name__].stop()
        if i > 0:
            _loop = digi.util.Loader(
                load_fn=fn,
                load_interval=i,
            )
            _report_loops[fn.__name__] = _loop
            _loop.start()


def name_from_auri(auri: tuple):
    return auri[3]


update = deep_set
get = deep_get



"""Views used for model manipulation."""
import os
import copy
from box import Box
from abc import ABC, abstractmethod
from kopf._cogs.structs.diffs import diff

import digi.util as util
from digi.util import deep_set

# XXX fix model and type view
class View(ABC):
    @abstractmethod
    def __init__(self,
                 root: dict,
                 root_key: str = "root",
                 gvr_str=None,
                 trim_gv: bool = True,
                 trim_mount: bool = True,
                 trim_name: bool = True,
                 ):
        self._src = root
        self._root = copy.deepcopy(self._src)
        self._root_key = root_key
        self._old, self._new = None, None

        self._trim_gv = trim_gv
        self._trim_mount = trim_mount
        self._trim_name = trim_name

        self._name = os.environ["NAME"]
        self._ns = os.environ.get("NAMESPACE", "default")

        if gvr_str is None:
            self._r = os.environ["PLURAL"]
            self._gv_str = f"{os.environ['GROUP']}" \
                           f"/{os.environ['VERSION']}"
            self._gvr_str = f"{self._gv_str}/{os.environ['PLURAL']}"
        else:
            gvr_str = util.full_gvr(gvr_str)
            self._r = util.parse_gvr(gvr_str)[-1]
            self._gv_str = "/".join(util.parse_gvr(gvr_str)[:-1])
            self._gvr_str = gvr_str

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, typ, value, traceback):
        raise NotImplementedError

    @abstractmethod
    def transform(self, root: dict, view: dict):
        raise NotImplementedError

    @abstractmethod
    def materialize(self) -> dict:
        raise NotImplementedError

    def m(self) -> dict:
        # shorthand
        return self.materialize()


class NameView(View):
    """
    Return all models in the current world/root view
    keyed by the namespaced name; if the nsn starts
    with default, it will be trimmed off; the original
    view is keyed by "root". Empty model without "spec"
    will be skipped.

    The __enter__ method constructs the model view from
    the root_view and __exit__ applies the changes back
    to the root_view.

    TBD: add mounts recursively but trim off mounts
      record the path and append on the diff path at exit;
      use _transform in __enter__
    """

    def __init__(self, *args, **kwargs):
        self._nsn_gvr = dict()
        super().__init__(*args, **kwargs)

    def __enter__(self):
        _view = {"root": self._root}
        _mount = self._root.get("mount", {})

        for typ, ms in _mount.items():
            for n, m in ms.items():
                if "spec" not in m:
                    continue
                n = util.trim_default_space(n)
                _view.update({n: m["spec"]})
                self._nsn_gvr[n] = typ

        self._old, self._new = _view, copy.deepcopy(_view)
        return self._new

    def __exit__(self, typ, value, traceback):
        _view = {"root": self._root}
        # diff and apply
        _src = self._src
        _diffs = diff(self._old, self._new)
        for op, path, old, new in _diffs:
            nsn = path[0]
            if nsn == "root":
                deep_set(_src, ".".join(path[1:]), new)
            else:
                typ = self._nsn_gvr[nsn]
                nsn = util.normalized_nsn(nsn)
                path = ["mount", typ, nsn, "spec"] + list(path[1:])
                deep_set(_src, path, new)

    def transform(self, root: dict, view: dict):
        _mount = root.get("mount", {})
        if self._trim_mount:
            trim_mount(root, trim_gv=self._trim_gv)
        for typ, models in _mount.items():
            for name, model in models.items():
                name = util.trim_default_space(name)
                if "spec" not in model:
                    continue

                spec: dict = model["spec"]
                self.transform(spec, view)
                view.update({name: spec})

    def materialize(self):
        root = self._root
        view = {self._root_key: root}
        self.transform(root, view)
        return view


class KindView(View):
    """
    Return models group-by their gvr, if the gv is
    the same as the parent's gv, it will be trimmed
    to only the plural.

    TBDs: ditto
    """

    def __init__(self, *args, **kwargs):
        self._typ_full_typ = dict()
        super().__init__(*args, **kwargs)

    def __enter__(self):
        # _view = {self._r: {"root": self._root_view}}
        _view = {"root": self._root}
        _mount = self._root.get("mount", {})

        for typ, ms in _mount.items():
            _typ = typ.replace(self._gv_str + "/", "") if self._trim_gv else typ
            _view[_typ] = {}
            self._typ_full_typ[_typ] = typ

            for n, m in ms.items():
                if "spec" not in m:
                    continue
                n = util.trim_default_space(n)
                _view[_typ].update({n: m["spec"]})

        self._old, self._new = _view, copy.deepcopy(_view)
        return self._new

    def __exit__(self, typ, value, traceback):
        _src = self._src
        _diffs = diff(self._old, self._new)

        for op, path, old, new in _diffs:
            typ = path[0]
            if typ == "root":
                deep_set(_src, ".".join(path[1:]), new)
            else:
                typ = self._typ_full_typ[typ]
                nsn = util.normalized_nsn(path[1])
                path = ["mount", typ, nsn, "spec"] + list(path[2:])
                deep_set(_src, path, new)

    def transform(self, root: dict, view: dict):
        _mount = root.get("mount", {})
        if self._trim_mount:
            trim_mount(root, trim_gv=self._trim_gv)
        for typ, models in _mount.items():
            _typ = typ.replace(self._gv_str + "/", "") if self._trim_gv else typ
            if _typ not in view:
                view[_typ] = list() if self._trim_name else dict()

            for name, model in models.items():
                name = util.trim_default_space(name)
                if "spec" not in model:
                    continue

                spec: dict = model["spec"]
                self.transform(spec, view)

                view[_typ].append(spec if self._trim_name else {name: spec})

    def materialize(self):
        root = self._root
        if self._trim_name:
            view = {self._root_key: [root]}
        else:
            view = {self._root_key: {self._name: root}}
        self.transform(root, view)
        return view


class CleanView(View):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self, typ, value, traceback):
        raise NotImplementedError

    def transform(self, root: dict, view: dict):
        _mount = root.get("mount", {})
        view.update(copy.deepcopy(root))
        if len(_mount) > 0:
            if self._trim_mount:
                trim_mount(view, trim_gv=self._trim_gv)
                return
            view["mount"] = dict()

        for typ, models in _mount.items():
            _typ = typ.replace(self._gv_str + "/", "")
            view["mount"][_typ] = dict()

            for name, model in models.items():
                new_name = util.trim_default_space(name)
                if "spec" not in model:
                    continue

                spec: dict = model["spec"]
                view["mount"][_typ][new_name] = dict()
                self.transform(spec, view["mount"][_typ][new_name])

    def materialize(self) -> dict:
        root, view = self._root, dict()
        self.transform(root, view)
        return view


class DotView(View):
    """Dot accessible models."""
    _char_map = {
        "-": "_",
        ".": "_",
        "/": "_",
        " ": "_",
        "\\": "_",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._src_view = self._src
        self._dot_view = None
        self._dot_view_old = None
        # map between unsafe attributes
        # to original ones
        self._attr_map = dict()

    def __enter__(self):
        # box does not record nor expose a conversion
        # map for the safe attributes, so we do so
        # ahead of time and pass a safe dict to box;
        # the self._attr_map keeps track of any conversion.
        self._dot_view_old = self._to_safe_dict(self._src_view)
        self._dot_view = Box(self._dot_view_old)
        return self._dot_view

    def __exit__(self, exc_type, exc_val, exc_tb):
        _src = self._src_view
        self._dot_view = self._dot_view.to_dict()
        _diffs = diff(self._dot_view_old, self._dot_view)
        for op, path, old, new in _diffs:
            path = [self._attr_map.get(p, p) for p in path]
            deep_set(_src, path, new)

    def transform(self, root: dict, view: dict):
        pass

    def materialize(self) -> dict:
        return self.__enter__()

    def _to_safe_dict(self, d: dict) -> dict:
        safe_d = dict()
        for k, v in d.items():
            orig_k = k
            k = self._to_safe_attr(k)
            self._attr_map[k] = orig_k
            if isinstance(v, dict):
                v = self._to_safe_dict(v)
            safe_d[k] = v
        return safe_d

    @staticmethod
    def _to_safe_attr(s: str):
        for k, v in DotView._char_map.items():
            s = s.replace(k, v)
        return s


# aliases
ModelView, TypeView = NameView, KindView


def trim_mount(v: dict, *,
               trim_all=False,
               trim_gv=True):
    # keep only the names of immediate children
    if not trim_all:
        for gvr, models in v.get("mount", {}).items():
            key = util.trim_gv(gvr) if trim_gv else gvr
            v[key] = list()
            for name, _ in models.items():
                name = util.trim_default_space(name)
                v[key].append(name)
    v.pop("mount", "")



import digi
import digi.data.sync as sync
import digi.data.util as util
from digi.data import logger, zed
from digi.data import flow as flow_lib


class Router:
    def __init__(self):
        self.ingress = Ingress()
        self.egress = Egress()


class Ingress:
    def __init__(self):
        self._syncs = dict()

    def start(self):
        for name, _sync in self._syncs.items():
            _sync.start()
            logger.info(f"started ingress sync {name} "
                        f"with query: {_sync.query_str}")

    def update(self, config: dict):
        self._syncs = dict()

        for name, ig in config.items():
            if ig.get("pause", False):
                continue

            sources = list()
            flow, flow_agg = ig.get("flow", ""), \
                             ig.get("flow_agg", "")
            for s in ig.get("source", []):
                sources += util.parse_source(s)
            for s in ig.get("sources", []):
                sources += util.parse_source(s)

            # TBD deduplicate sources
            if len(sources) == 0:
                continue

            if flow_agg == "":
                _out_flow = flow_lib.refresh_ts
            else:
                _out_flow = f"{flow_agg} | {flow_lib.refresh_ts}"

            _sync = sync.Sync(
                sources=sources,
                in_flow=flow,
                out_flow=_out_flow,
                dest=digi.pool.name,
                eoio=ig.get("eoio", True),
                patch_source=ig.get("patch_source", False),
                client=zed.Client(),
                owner=digi.name,
            )
            self._syncs[name] = _sync

    def stop(self):
        for _, _sync in self._syncs.items():
            _sync.stop()

    def restart(self, config: dict):
        self.stop()
        self.update(config=config)
        self.start()

    @staticmethod
    def _any_source():
        # TBD fetch all sources - digi and egresses
        # for digis that are mounted
        raise NotImplementedError


class Egress:
    INIT = {"__meta": "init"}

    def __init__(self):
        self._syncs = dict()

    def start(self):
        for name, _sync in self._syncs.items():
            _sync.start()
            logger.info(f"started egress sync {name} "
                        f"with query:\n{_sync.query_str}")

    def update(self, config: dict):
        self._syncs = dict()

        for name, ig in config.items():
            if ig.get("driver_managed", False) \
                    or ig.get("pause", False):
                continue

            flow = ig.get("flow", "")
            _sync = sync.Sync(
                sources=[digi.pool.name],
                in_flow=flow,
                out_flow=f"{flow_lib.drop_meta} | {flow_lib.refresh_ts}",
                dest=f"{digi.pool.name}@{name}",
                eoio=ig.get("eoio", True),
                client=zed.Client(),
                owner=digi.name,
            )
            self._syncs[name] = _sync
            # TBD garbage collect unused branches
            digi.pool.create_branch_if_not_exist(name)

    def stop(self):
        for _, _sync in self._syncs.items():
            _sync.stop()

    def restart(self, config: dict):
        self.stop()
        self.update(config=config)
        self.start()


@digi.on.mount
def do_mount(model, diff):
    config = model.get("ingress", {})
    # TBD move new_mount to decorator
    # TBD filter to relevant mounts only
    # XXX mount removal doesn't provide new data
    if digi.on.mount_change(diff):
        digi.router.ingress.restart(config)


@digi.on.ingress(prio=128)  # handler runs at last XXX sys.maxsize
def do_ingress(config):
    digi.router.ingress.restart(config)


@digi.on.egress(prio=128)
def do_egress(config):
    digi.router.egress.restart(config)


def create_router():
    return Router()



import os
import logging

# default logger
logger = logging.getLogger(os.environ.get("LOGGER_NAME", __name__))
log_level = int(os.environ.get("LOGLEVEL", logging.INFO))
logger.setLevel(log_level)

# digi metadata and configurations
g = group = os.environ.get("GROUP", "digi.dev")
v = version = os.environ.get("VERSION", "v1")
k = kind = os.environ.get("KIND", "test")
r = resource = os.environ.get("PLURAL", "tests")
n = name = os.environ.get("NAME", "t1")
ns = namespace = os.environ.get("NAMESPACE", "default")
duri = auri = (g, v, r, n, ns)

pool_provider = os.environ.get("POOL_PROVIDER", "zed")
load_trim_mount = os.environ.get("TRIM_MOUNT_ON_LOAD", "") != "false"
enable_mounter = os.environ.get("MOUNTER", "") == "true"
enable_visual = os.environ.get("VISUAL", "") == "true"
visual_type = os.environ.get("VISUAL_TYPE", "")
visual_refresh_interval = float(os.environ.get("VISUAL_REFRESH_INTERVAL", 1000))

# digi modules; force init
from digi import (
    on,
    util,
    mount,
    filter,
    view,
    data,
    control,
    dbox,
)
from digi.main import run
from digi.reconcile import rc
model, pool, router = None, None, None

__all__ = [
    "on", "util", "view", "filter",
    "run", "logger", "mount", "rc",
    "model", "pool", "router", "dbox",
    "data", "control"
]



import digi
from digi import on, util, filter
import typing
import random

"""
dbox provides utilities for handling a scene hierarchy:
- set random seed using meta.seed
- set children's meta.managed
- filter ``managed'' for handler event filtering in children;
  e.g. @on.mount(cond=dbox.managed)
- event generation loop. Generation interval controlled by gen_interval,
  default to -1 which pauses the event generation; 
  e.g. @dbox.loop(managed=False)
dbox uses its own global random generator instance to separate from
other use cases of random module.
"""
seed = 42
event_random = random.Random()
sim_random = random.Random()
random = random.Random()


def init_default():
    """This adds the following handlers:
    - Handler that updates random.seed based on meta.seed
    - Handler that configures all mounts to managed. """
    seeding()
    managing()


init = init_default


def seeding():
    global seed, random, event_random, sim_random
    cur_seed = seed
    random.seed(cur_seed)
    event_random.seed(cur_seed)
    sim_random.seed(cur_seed)

    @on.meta
    def do_seed(meta):
        global seed
        nonlocal cur_seed
        _seed = meta.get("seed")
        if _seed != cur_seed:
            cur_seed = _seed
            if cur_seed is not None:
                random.seed(_seed)
                seed = _seed


def managing():
    # TBD mount filter on new children (skip other updates)
    # TBD model class that contains advanced/processed events
    @on.mount
    def do_manage(mounts):
        manage(mounts)


def manage(mounts, skip_gvr=None):
    if skip_gvr is None:
        skip_gvr = {}
    for gvr, mocks in mounts.items():
        if gvr in skip_gvr:
            continue
        for _, mock in mocks.items():
            util.update(mock, "spec.meta.managed", True)


def managed(proc_spec, diff, *args, **kwargs):
    """handler filter decides when to skip reconciliation.
    Used as e.g. @on.mount(cond=dbox.managed)"""
    _, _ = args, kwargs
    managed = util.get(proc_spec, "meta.managed", False)
    if managed:
        return False
    else:
        return filter.path_changed(diff, ("meta", "managed")) or \
               filter.changed(proc_spec, diff, *args, **kwargs)


_loops = dict()


def loop(fn: typing.Callable, managed=True):
    """E.g. @dbox.loop(managed=False)"""
    # TBD take interval function
    global _loops
    _loops[fn.__name__] = util.Loop(loop_fn=fn)

    @on.meta
    def do_loop(meta):
        nonlocal managed
        i, _managed = meta.get("gen_interval", -1), \
                      managed and meta.get("managed", False)
        _loops[fn.__name__].stop()
        if i > 0 and not _managed:
            _loop = digi.util.Loop(
                loop_fn=fn,
                loop_interval=i,
            )
            _loops[fn.__name__] = _loop
            _loop.start()


def make_event_interval(avg_t):
    def fn() -> int:
        min_, max_ = int(avg_t / 2), int(avg_t * 2)
        return random.randint(min_, max_)

    return fn



import os
import logging

logger = logging.getLogger(__name__)
os.environ["ZED_LAKE"] = os.environ.get("ZED_LAKE", "http://lake:6534")
lake_url = os.environ["ZED_LAKE"]

from digi.data.pool import create_pool
from digi.data.router import create_router
from digi.data.zed import Client
from digi.data.sync import Sync, Watch

lake = Client()

__all__ = [
    "Sync", "Watch",
    "create_router", "create_pool"
]



import typing
import datetime
import dateutil.parser
import durationpy
import binascii
import decimal
import ipaddress
import json

"""TBD patch upstream zed/zed.py"""

_py_to_zed_primitive_type = {
    "<class 'int'>": "int64",
    "<class 'datetime.timedelta'>": "duration",
    "<class 'datetime.datetime'>": "time",
    "<class 'float'>": "float64",
    "<class 'decimal.Decimal'>": "decimal",
    "<class 'bool'>": "bool",  # TBD check zed/zed.py 'bool' -> 'T'
    "<class 'bytes'>": "bytes",
    "<class 'str'>": "string",
    "<class 'ipaddress.IPv4Address'>": "ip",
    "<class 'ipaddress.IPv4Network'>": "net",
    "<class 'type'>": "type",
    "<class 'NoneType'>": "null",
}


class __Counter:
    def __init__(self):
        self.ctr = 29

    def get_and_inc(self):
        self.ctr += 1
        return self.ctr


def encode(objs: typing.List[dict]) -> str:
    id_ctr, types = __Counter(), dict()
    for obj in objs:
        yield json.dumps({
            "type": _encode_type(id_ctr, obj),
            "value": _encode_value(obj),
        })


def _encode_type(id_ctr, value) -> dict:
    # TBD ref type
    if isinstance(value, dict):
        return {
            "kind": "record",
            "fields": [
                {"name": _f, "type": _encode_type(id_ctr, _v)}
                for _f, _v in value.items()
            ] if len(value) > 0 else None,
            "id": id_ctr.get_and_inc()
        }
    elif isinstance(value, list):
        return {
            "kind": "array",
            "type": _encode_type(id_ctr, value[0] if len(value) > 0 else None),
            "id": id_ctr.get_and_inc()
        }
    elif isinstance(value, set):
        return {
            "kind": "set",
            "type": _encode_type(id_ctr, value[0] if len(value) > 0 else None),
            "id": id_ctr.get_and_inc()
        }
    else:  # primitive type
        typ = str(type(value))
        if typ not in _py_to_zed_primitive_type:
            raise Exception(f"unknown zed primitive type for {typ}")
        return {
            "kind": "primitive",
            "name": _py_to_zed_primitive_type[typ],
        }


def _encode_value(value) -> typing.Union[list, str, None]:
    if isinstance(value, dict):
        return [_encode_value(_v) for _, _v in value.items()]
    elif isinstance(value, list):
        return [_encode_value(_v) for _v in value]
    elif isinstance(value, set):
        return [_encode_value(_v) for _v in value]
    elif isinstance(value, datetime.datetime):
        return encode_datetime(value)
    elif isinstance(value, datetime.timedelta):
        return durationpy.to_str(value)
    elif str(type(value)) in _py_to_zed_primitive_type:
        if value is None:
            return None
        if type(value) == bool:
            return str(value).lower()
        else:
            return str(value)
    else:
        raise Exception(f"cannot encode value of type {type(value)}")


def encode_datetime(value: datetime.datetime) -> str:
    return value.isoformat().replace("+00:00", "") + "Z"


class QueryError(Exception):
    """Raised by Client.query() when a query fails."""
    pass


def decode_raw(raw):
    types = {}
    for msg in raw:
        typ, value = msg['type'], msg['value']
        if isinstance(typ, dict):
            yield _decode_value(_decode_type(types, typ), value)
        elif typ == 'QueryError':
            raise QueryError(value['error'])


def _decode_type(types, typ):
    kind = typ['kind']
    if kind == 'ref':
        return types[typ['id']]
    if kind == 'primitive':
        return typ
    elif kind == 'record':
        if typ['fields'] is None:
            # XXX
            pass
        for f in typ['fields']:
            f['type'] = _decode_type(types, f['type'])
    elif kind in ['array', 'set']:
        typ['type'] = _decode_type(types, typ['type'])
    elif kind == 'map':
        typ['key_type'] = _decode_type(types, typ['key_type'])
        typ['val_type'] = _decode_type(types, typ['val_type'])
    elif kind == 'union':
        typ['types'] = [_decode_type(types, t) for t in typ['types']]
    elif kind == 'enum':
        pass
    elif kind in ['error', 'named']:
        typ['type'] = _decode_type(types, typ['type'])
    else:
        raise Exception(f'unknown type kind {kind}')
    types[typ['id']] = typ
    return typ


def _decode_value(typ, value):
    if value is None:
        return None
    kind = typ['kind']
    if kind == 'primitive':
        name = typ['name']
        if name in ['uint8', 'uint16', 'uint32', 'uint64',
                    'int8', 'int16', 'int32', 'int64']:
            return int(value)
        if name == 'duration':
            return durationpy.from_str(value)
        if name == 'time':
            return dateutil.parser.isoparse(value)
        if name in ['float16', 'float32', 'float64']:
            return float(value)
        if name == 'decimal':
            return decimal.Decimal(value)
        if name == 'bool':
            return value in {'T', 'true'}  # TBD patch upstream
        if name == 'bytes':
            return binascii.a2b_hex(value[2:])
        if name == 'string':
            return value
        if name == 'ip':
            return ipaddress.ip_address(value)
        if name == 'net':
            return ipaddress.ip_network(value)
        if name in 'type':
            return value
        if name == 'null':
            return None
        raise Exception(f'unknown primitive name {name}')
    if kind == 'record':
        return {f['name']: _decode_value(f['type'], v)
                for f, v in zip(typ['fields'], value)}
    if kind == 'array':
        return [_decode_value(typ['type'], v) for v in value]
    if kind == 'set':
        return {_decode_value(typ['type'], v) for v in value}
    if kind == 'map':
        key_type, val_type = typ['key_type'], typ['val_type']
        return {_decode_value(key_type, v[0]): _decode_value(val_type, v[1])
                for v in value}
    if kind == 'union':
        type_index, val = value
        return _decode_value(typ['types'][int(type_index)], val)
    if kind == 'enum':
        return typ['symbols'][int(value)]
    if kind in ['error', 'named']:
        return _decode_value(typ['type'], value)
    raise Exception(f'unknown type kind {kind}')


if __name__ == '__main__':
    test = [
        12,
        {},
        [],
        {"watt": "12"},
        {"watt": "12", "user": {"alice": True}},
    ]
    print("\n".join((encode(test))))
    # import zed
    # TBD fix zed.py/decoder handle null fields
    # print(list(zed.decode_raw(json.loads(line) for line in (encode(test)))))



import os
import time
import datetime
import threading
import typing
import requests
import json
import yaml
from collections import defaultdict

from . import zed, zjson  # TBD use upstream

default_lake_url = os.environ.get("ZED_LAKE", "http://localhost:9867")


class Sync(threading.Thread):
    """Many-to-one sync between data pools on Zed lake."""
    SOURCE_COMMIT = 1
    SKIP = 2

    def __init__(self,
                 sources: typing.List[str],
                 dest: str,
                 in_flow: str = "",
                 out_flow: str = "",
                 *,
                 poll_interval: float = -1,  # sec, <0: use push
                 eoio: bool = True,  # exactly-once in-order
                 patch_source: bool = False,
                 owner: str = "sync",  # commit author
                 lake_url: str = default_lake_url,
                 client: zed.Client = None,
                 ):
        assert len(sources) > 0 and dest != ""
        self.sources = self._normalize(sources)
        self.dest = self._normalize_one(dest)
        self.in_flow = in_flow  # TBD multi-in_flow
        self.out_flow = out_flow
        self.poll_interval = poll_interval
        self.owner = owner
        self.patch_source = patch_source
        # if eoio enabled, the sync agent will
        # process only those records that contain
        # a 'ts' field; XXX assume pool key is ts
        self.eoio = eoio
        self.client = zed.Client(base_url=lake_url) if client is None else client
        self.source_ts = self._fetch_source_ts()  # track {source: max(ts)}
        self.source_pool_ids = self._fetch_source_pool_ids()
        self.source_set = set(self.sources)
        self.query_str = self._make_query()

        threading.Thread.__init__(self)
        self._stop_flag = threading.Event()

    def run(self):
        self._stop_flag.clear()

        self.once()
        if self.poll_interval > 0:
            self._poll_loop()
        else:
            self._event_loop()

    def stop(self):
        self._stop_flag.set()

    def once(self):
        records = self.read()
        if len(records) != 0:
            self.load(records)

    def read(self) -> list:
        records = list()
        if self.eoio:
            self.query_str = self._make_query()
        for r in self.client.query(self.query_str):
            if "__from" not in r:
                records.append(r)
                continue
            source, max_ts = r["__from"], r["max_ts"]
            if self.eoio:
                if max_ts is None:
                    raise Exception(f"no ts found in records from {source}")
                if source not in self.source_ts:
                    self.source_ts[source] = max_ts
                else:
                    self.source_ts[source] = max(max_ts, self.source_ts[source])
            else:
                self.source_ts[source] = datetime.datetime.min
        return records

    def load(self, records: list):
        records = "\n".join(zjson.encode(records))
        dest_pool, dest_branch = self._denormalize_one(self.dest)
        self.client.load(
            dest_pool, records,
            branch_name=dest_branch,
            commit_author=self.owner,
            meta=self._source_ts_json(),
        )

    def _event_loop(self):
        s = requests.Session()
        with s.get(f"{default_lake_url}/events",
                   headers=None, stream=True) as resp:
            lines = resp.iter_lines()
            for line in lines:
                if self._stop_flag.is_set():
                    return
                event = self._parse_event(line, lines)
                if event == Sync.SOURCE_COMMIT:
                    self.once()
                elif event == Sync.SKIP:
                    continue
                else:
                    raise NotImplementedError

    def _poll_loop(self):
        while not self._stop_flag.is_set():
            self.once()
            time.sleep(self.poll_interval)

    def _make_query(self) -> str:
        in_str = "from (\n"
        for source in self.sources:
            cur_ts = self.source_ts.get(source, datetime.datetime.min)
            filter_flow = f"ts > {zjson.encode_datetime(cur_ts)} |" \
                if self.eoio else ""
            patch_source_flow = f"put from := '{source}' |" \
                if self.patch_source else ""
            in_str += f"pool {source} => {filter_flow} {patch_source_flow} fork (" \
                      f"=> select max(ts) as max_ts | put __from := '{source}' " \
                      f"=> {'pass' if self.in_flow == '' else self.in_flow})"
            if len(self.sources) > 1:
                in_str += "\n"
        in_str += ")\n"  # wrap up from clause
        out_str = f"switch (case has(__from) => pass default => " \
                  f"{'pass' if self.out_flow == '' else self.out_flow})"
        return f"{in_str} | sort this | {out_str}"

    def _source_ts_json(self) -> str:
        return json.dumps({
            source: zjson.encode_datetime(ts)
            for source, ts in self.source_ts.items()
        })

    def _fetch_source_ts(self) -> dict:
        source_ts = dict()
        pool, branch = self._denormalize_one(self.dest)
        # filter to branches that have at least one commit
        branch_flow = f"from {pool}:branches | " \
                      f"select branch.name as branch," \
                      f"branch.commit as commit |" \
                      f"commit > 0x{'0' * 40}"
        if branch not in set(
                r["branch"] for r in self.client.query(branch_flow)
        ):
            return source_ts
        meta_flow = f"from {self.dest}:log | " \
                    f"typeof(this)==<Commit> | " \
                    f"over meta | " \
                    f"max(cast(value, <time>)) by key | " \
                    f"rename max_ts:=max"
        for r in self.client.query(meta_flow):
            source_ts[r["key"][0]] = r["max_ts"]
        return source_ts

    def _fetch_source_pool_ids(self) -> dict:
        return {
            f"0x{r['id'].hex()}": r["name"]
            for r in self.client.query("from :pools")
            if r["name"] in set(s.split("@")[0] for s in self.sources)
        }

    def _parse_event(self, line: bytes, lines: typing.Iterator) -> int:
        def substr(s, start, end):
            return (s.split(start))[1].split(end)[0]

        line = line.decode()
        if line == "event: branch-commit":
            data = next(lines).decode().lstrip("data: ")
            pool_id = substr(data, "pool_id:", ",")
            branch = substr(data, "branch:", ",").strip('"')
            # TBD cache commit_id
            # commit_id = substr(data, "commit_id:", ",")
            if pool_id in self.source_pool_ids:
                pool_name = self.source_pool_ids[pool_id]
                if f"{pool_name}@{branch}" in self.source_set:
                    return Sync.SOURCE_COMMIT
        return Sync.SKIP

    @staticmethod
    def _normalize(names: list) -> list:
        return [Sync._normalize_one(n) for n in names]

    @staticmethod
    def _normalize_one(name: str) -> str:
        """Return source in form of pool@main."""
        return f"{name}@main" if "@" not in name else name

    @staticmethod
    def _denormalize_one(name: str) -> tuple:
        pool, branch = name.split("@")
        return pool, branch


class Watch(Sync):
    """A destination-less sync that runs a UDF in once()."""

    def __init__(self, fn: typing.Callable,
                 source_ts=None, *args, **kwargs):
        self.fn = fn
        self.source_ts = source_ts

        super().__init__(dest="none", *args, **kwargs)

    def once(self):
        self.fn(self.read())

    def _fetch_source_ts(self) -> dict:
        if self.source_ts is None:
            return defaultdict(lambda: datetime.datetime.min)
        else:
            return self.source_ts


def from_config(path: str) -> Sync:
    with open(path) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return Sync(
        sources=config["sources"],
        dest=config["dest"],
        in_flow=config.get("in_flow", ""),
        out_flow=config.get("out_flow", ""),
        poll_interval=config.get("poll_interval", -1),
    )


if __name__ == '__main__':
    from_config("demo.yaml").start()



import os
import yaml
import copy
import pprint as pp

from digi.view import (
    ModelView,
    TypeView,
    DotView,
    CleanView,
)


# TBD pytest
def test():
    os.environ.update({
        "GROUP": "mock.digi.dev",
        "VERSION": "v1",
        "PLURAL": "rooms",
        "NAME": "room",
        "NAMESPACE": "default",
    })
    test_yaml = """
    control:
      brightness:
        intent: 0.8
        status: 0
      mode:
        intent: sleep
        status: sleep
    mount:
      mock.digi.dev/v1/lamps:
        default/lamp-test:
          spec:
            control:
              power: 
                intent: "on"
    """

    # TBD root tests
    orig_v = yaml.load(test_yaml, Loader=yaml.FullLoader)
    v = copy.deepcopy(orig_v)
    print("# Using model view, before:", v)
    with ModelView(v) as mv:
        mv["lamp-test"]["control"]["power"]["intent"] = "off"
        mv["root"]["control"]["mode"]["status"] = "away"
    print(f"-----\nafter: {v}\n")

    v = copy.deepcopy(orig_v)
    print("# Using type view, before:", v)
    with TypeView(v) as tv:
        tv["lamps"]["lamp-test"]["control"]["power"]["intent"] = "off"
        mv["root"]["control"]["mode"]["status"] = "away"
    print(f"-----\nafter: {v}\n")

    v = copy.deepcopy(orig_v)
    print("with type + dot view chain, before:", v)
    with TypeView(v) as tv, DotView(tv) as dv:
        dv.lamps.lamp_test.control.power.intent = "off"
        dv.root.control.brightness.status = "away"
    print(f"-----\nafter: {v}\n")

    v = copy.deepcopy(orig_v)
    print("flattened model view, before:", v)
    print(f"-----\nafter: {ModelView(v, 'room').m()}\n")

    v = copy.deepcopy(orig_v)
    print("flattened type view, before:", v)
    print(f"-----\nafter: {TypeView(v, 'rooms').m()}\n")

    v = copy.deepcopy(orig_v)
    print("clean view, before:", v)
    print(f"-----\nafter: {CleanView(v, 'rooms').m()}\n")

if __name__ == '__main__':
    test()



"""Logic and policy processors"""

# import pyjq
import time
from digi import logger
from digi.view import ModelView


def jq(policy: str):
    raise NotImplementedError
    # preproc macro
    # _macros = {
    #     "$time": str(time.time()),
    #     # ...
    # }
    #
    # for _m, _v in _macros.items():
    #     policy = policy.replace(_m, _v)
    #
    # logger.info(f"processor: jq running policy {policy}")
    #
    # def fn(proc_view, *args, **kwargs):
    #     _, _ = args, kwargs
    #     with ModelView(proc_view) as mv:
    #         mv.update(pyjq.one(policy, mv))
    #
    # return fn


def py(policy: str):
    return NotImplemented


"""Event filters"""


def always(*args, **kwargs):
    _, _ = args, kwargs
    return True


def changed(_, diff, path, *args, **kwargs) -> bool:
    _, _ = args, kwargs
    # TBD: add shared diff for all handlers
    # TBD: support incremental diff
    return path_changed(diff, path)


def path_changed(diff: list, path: tuple):
    changed_paths = {(".",): True}
    for op, path_, old, new in diff:
        # when create
        if old is None and len(path_) == 0:
            changed_paths.update(_from_model(new))
        else:
            changed_paths.update(_from_path_tuple(path_))
    if path in changed_paths or len(diff) == 0:
        return True
    return False


def _from_model(d: dict):
    result = dict()
    to_visit = [[d.get("spec", {}), []]]
    for n, prefix in to_visit:
        result[tuple(prefix)] = True
        if type(n) is not dict:
            continue
        for _k, _v in n.items():
            to_visit.append([_v, prefix + [_k]])
    return result


def _from_path_tuple(p: tuple):
    # expand a path tuple to dict of paths
    return {
        # skip "spec"
        p[1:_i + 1]: True
        for _i in range(len(p))
    }



import os
import sys
import subprocess
import time
import yaml
import inflection
import digi.util as util

_dir = os.path.dirname(os.path.realpath(__file__))
_mock_dir = os.path.join(_dir, "..", "mocks")
_parent = os.path.join(_mock_dir, "room")
_child = os.path.join(_mock_dir, "lamp")
_parent_cr = os.path.join(_parent, "test", "cr.yaml")
_child_cr = os.path.join(_child, "test", "cr.yaml")
_parent_crd = os.path.join(_parent, "crd.yaml")
_child_crd = os.path.join(_child, "crd.yaml")


def _wait(t=0.5):
    time.sleep(t)


"""
Mount tests:

Create/resume/delete events:
1. create parent model; neat display parent -> parent
2. create child model; neat display child -> child
3. dq mount child parent; neat display parent and child:
    - parent.mount.child
    - parent.mount.child.status matches child.status
    - child.intent matches parent.mount.child.intent
4. dq mount -d child parent; neat display parent and child:
    - parent.mount.child is gone
5. dq mount child parent; kc remove child model; neat display parent
    - parent.mount.child is gone 
6. create child model; neat display parent and child
    - same as 3.

Update events:
1. create parent and child model
2. dq mount child parent
3. update parent.mount.child.intent
    - child.intent matches 
4. update parent.mount.child.status
    - child.status doesn't match
5. update child.status  
    - parent.mount.child.status matches
6. update child.intent
    - parent.mount.child.intent matches
"""


def test_mount(parent, child):
    _make_cr(parent)
    _show(parent, msg="created parent")

    _make_cr(child)
    _show(child, msg="created child")

    _wait()
    _mount(child, parent)

    _wait()
    _show(parent, msg="mounted child to parent")

    _wait(60)
    _mount(child, parent, unmount=True)
    _show(parent, msg="unmounted child")

    _wait(5)
    _mount(child, parent)
    _wait()
    _make_cr(child, delete=True)
    _show(parent, msg="mount child but delete it after")

    _wait()
    _make_cr(child)
    _show(parent, msg="create child again")


def test_prop(parent, child):
    _make_cr(parent)
    _show(parent, msg="created parent")

    _make_cr(child)
    _show(child, msg="created child")

    _wait()
    _mount(child, parent)

    _wait()
    _show(parent, msg="mounted child to parent")

    ps, cs = _get_spec(parent), _get_spec(child)
    for _, ms in ps["spec"]["mount"].items():
        for _, m in ms.items():
            print(m)


# TBD move to pytest
def test_all():
    with _Setup() as s:
        test_mount(s.parent, s.child)

    # with _Setup() as s:
    #     ...
    # test_prop(s.parent, s.child)


def _get_spec(m):
    return util.get_spec(m["g"], m["v"], m["r"],
                         m["n"], m["ns"])


class _Setup:
    @staticmethod
    def _parse(cr):
        with open(cr) as f:
            model = yaml.load(f, Loader=yaml.SafeLoader)
            meta = model["metadata"]
            auri = "/".join(["", model["apiVersion"], model["kind"],
                             meta.get("namespace", "default"),
                             meta["name"]])
            g, v = model["apiVersion"].split("/")
            return {
                "auri": auri,
                "g": g,
                "v": v,
                "k": model["kind"].lower(),
                "n": meta["name"],
                "ns": meta.get("namespace", "default"),
                "r": inflection.pluralize(model["kind"]).lower(),
                "alias": meta["name"],
                "cr_file": cr,
            }

    def __enter__(self):
        _apply(_parent_crd)
        self.parent = self._parse(_parent_cr)
        _apply(_child_crd)
        self.child = self._parse(_child_cr)

        _make_cr(self.parent, delete=True)
        _make_cr(self.child, delete=True)

        _cmd(f"dq alias {self.parent['auri']} {self.parent['alias']}")
        _cmd(f"dq alias {self.child['auri']} {self.child['alias']}")

        _make_driver(self.parent["k"], self.parent["n"], delete=True)
        _make_driver(self.parent["k"], self.parent["n"])

        print("waiting for driver ready...")
        _wait(5)
        return self

    def __exit__(self, *args, **kwargs):
        _make_driver(self.parent["k"], self.parent["n"], delete=True)

        _make_cr(self.parent, delete=True)
        _apply(_parent_crd, delete=True)

        _make_cr(self.child, delete=True)
        _apply(_child_crd, delete=True)


def _mount(child, parent, unmount=False):
    _cmd(f"dq mount {'-d ' if unmount else ''}"
         f"{child['alias']} {parent['alias']}",
         show_cmd=True)


def _make_driver(kind, name, delete=False):
    _cmd(f"cd {_mock_dir}; "
         f"dq {'stop' if delete else 'run'} {kind} {name} | true")


def _make_cr(model, delete=False):
    _apply(model["cr_file"], delete=delete)


def _apply(f, delete=False):
    _cmd(f"kubectl {'delete' if delete else 'apply'} "
         f"-f {f} {'> /dev/null 2>&1 | true' if delete else ''}")


def _show(model, neat=True, msg=""):
    print("\n---" + msg)
    _cmd(f"kubectl get {model['k']} {model['n']} "
         f"-oyaml {'| kubectl neat' if neat else ''}")


def _cmd(cmd, show_cmd=False, quiet=False):
    """Executes a subprocess running a shell command and
    returns the output."""
    if show_cmd:
        print()
        print("$", cmd)

    if quiet:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            executable='/bin/bash')
    else:
        proc = subprocess.Popen(cmd, shell=True,
                                executable='/bin/bash')

    out, _ = proc.communicate()

    if proc.returncode:
        if quiet:
            print('Log:\n', out, file=sys.stderr)
        print('Error has occurred running command: %s' % cmd,
              file=sys.stderr)
        sys.exit(proc.returncode)
    return out


if __name__ == '__main__':
    test_all()



import typing
import inspect
from collections import OrderedDict

import digi
import digi.util as util
from digi.reconcile import rc

"""Filters."""


def meta(*args, **kwargs):
    # if decorator not parameterized
    if len(args) >= 1 and callable(args[0]):
        register(path="meta", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="meta." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="meta." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="meta", *args, **kwargs)
        return fn

    return decorator


def control(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="control", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="control." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="control." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="control", *args, **kwargs)
        return fn

    return decorator


def data(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="data", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="data." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="data." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="data", *args, **kwargs)
        return fn

    return decorator


def obs(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="obs", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="obs." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="obs." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="obs", *args, **kwargs)
        return fn

    return decorator


# TODO fix dbox compatibility
status = obs


def mount(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="mount", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="mount." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="mount." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="mount", *args, **kwargs)
        return fn

    return decorator


def ingress(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="ingress", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="ingress." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="ingress." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="ingress", *args, **kwargs)
        return fn

    return decorator


def egress(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path="egress", *args, **kwargs)
        return args[0]

    def decorator(fn):
        if len(args) >= 1:
            register(fn, path="egress." + args[0], *args[1:], **kwargs)
        elif "path" in kwargs:
            register(fn, path="egress." + kwargs.pop("path"), *args, **kwargs)
        else:
            register(fn, path="egress", *args, **kwargs)
        return fn

    return decorator


def model(*args, **kwargs):
    if len(args) >= 1 and callable(args[0]):
        register(path=".", *args, **kwargs)
        return args[0]

    def decorator(fn):
        register(fn, *args, **kwargs)
        return fn

    return decorator


def register(fn, path=".", prio=0, cond=digi.filter.changed):
    # preprocess the path str -> tuple of str
    _path = list()
    ps = path.split(".")

    child_typ = None

    if path == ".":
        _path = ["."]
    elif ps[0] == "mount":
        # XXX better . operator handling; use regex
        ps_gvr = path.split("/")
        assert len(ps_gvr) == 1 or len(ps_gvr) == 3
        if len(ps) == 1:
            # @mount w/o parameters
            _path = ps
        elif len(ps_gvr) == 1:
            # this gvr does not have group and version
            gvr = util.gvr(rc.g, rc.v, ps[1])
            ps[1], child_typ = gvr, gvr
            _path = ps
        elif len(ps_gvr) == 3:
            # TBD handle dot in version
            gvr = util.gvr(ps_gvr[0].replace("mount.", ""),
                           ps_gvr[1],
                           ps_gvr[2].split(".")[0])
            child_typ = gvr
            _path = ["mount", gvr] + ps_gvr[2].split(".")[1:]
    else:
        _path = ps
    # XXX assume default gv in gvr until fix dot in path literal;
    # _path = path.split(".")

    # TBD: join multiple path to allow multiple decorators per handler
    _path = tuple(_path)

    sig = inspect.signature(fn)

    # allow the handler declaration to omit arguments
    # the handler takes in argument in form of [subview, view, old_view]
    kwarg_filter = dict()
    args = OrderedDict(sig.parameters)

    # allow aliases
    for p in ["subview", "sub_view", "sv"]:
        if p in sig.parameters:
            kwarg_filter.update({"subview": p})
            args[p] = None

    for p in ["proc_view", "pv", "cur", "parent", "root", "model"]:
        if p in sig.parameters:
            kwarg_filter.update({"proc_view": p})
            args[p] = None

    for p in ["view", "v"]:
        if p in sig.parameters:
            kwarg_filter.update({"view": p})
            args[p] = None

    for p in ["old_view", "ov", "o"]:
        if p in sig.parameters:
            kwarg_filter.update({"old_view": p})
            args[p] = None

    for p in ["mount", "mounts", "mt", "mts",
              "child", "children"]:
        if p in sig.parameters:
            kwarg_filter.update({"mount": p})
            args[p] = None

    for p in ["obs"]:
        if p in sig.parameters:
            kwarg_filter.update({"obs": p})
            args[p] = None

    for p in ["back_prop", "bp"]:
        if p in sig.parameters:
            kwarg_filter.update({"back_prop": p})
            args[p] = None

    for p in ["diff"]:
        if p in sig.parameters:
            kwarg_filter.update({"diff": p})
            args[p] = None

    for p in ["meta"]:
        if p in sig.parameters:
            kwarg_filter.update({"meta": p})
            args[p] = None

    for p in ["typ", "child_typ"]:
        if p in sig.parameters:
            kwarg_filter.update({"typ": p})
            args[p] = None

    for i, (k, v) in enumerate(args.items()):
        if v is None:
            continue
        if i == 0:
            kwarg_filter["subview"] = k
        elif i == 1:
            kwarg_filter["proc_view"] = k
        elif i == 2:
            kwarg_filter["view"] = k
        elif i == 3:
            kwarg_filter["old_view"] = k
        elif i == 4:
            kwarg_filter["mount"] = k
        elif i == 5:
            kwarg_filter["obs"] = k
        elif i == 6:
            kwarg_filter["back_prop"] = k
        elif i == 7:
            kwarg_filter["diff"] = k
        elif i == 8:
            kwarg_filter["meta"] = k
        else:
            break

    def wrapper_fn(subview, proc_view, view,
                   old_view, mount, obs, back_prop,
                   diff, meta):
        kwargs = dict()
        for _k, _v in [("subview", subview),
                       ("proc_view", proc_view),
                       ("view", view),
                       ("old_view", old_view),
                       ("mount", mount),
                       ("obs", obs),
                       ("back_prop", back_prop),
                       ("diff", diff),
                       ("meta", meta),
                       ("typ", child_typ),
                       ]:
            if _k in kwarg_filter:
                kwargs[kwarg_filter[_k]] = _v
        fn(**kwargs)

    rc.add(handler=wrapper_fn,
           priority=prio,
           condition=cond,
           path=_path)


def mount_change(diff, gvr=None) -> bool:
    """Detect whether the diffs contain newly mounted digi."""
    for _diff in diff:
        op, path, _, _ = _diff
        # XXX simplify conditions
        if op == "remove" and len(path) == 4 and path[:2] == ("spec", "mount"):
            return True
        if op != "add" or (len(path) > 0 and path[-1] != "generation"):
            continue
        if gvr is None and path[:2] == ("spec", "mount"):
            return True
        elif path[:3] == ("spec", "mount", gvr):
            return True
    return False


def pool(*args, **kwargs):
    # TBD load max_ts on restart; the max_ts should be stored
    #   at a persistent location and written atomically, e.g.,
    #   the digi.model or digi.pool

    class DelayedWatch:
        def __init__(self, fn, *args, **kwargs):
            self.fn = fn
            self.args = args
            self.kwargs = kwargs
            self.watch = None

        def start(self):
            self.watch = digi.pool.watch(
                self.fn, *self.args, **self.kwargs
            )
            self.watch.start()

        def stop(self):
            self.watch.stop()

    # @pool
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        digi.rc.add_data_watch(
            name=watch_name(args[0]),
            watch=DelayedWatch(args[0], *args, **kwargs),
        )
        return args[0]

    # @pool(*args, **kwargs)
    def decorator(fn):
        digi.rc.add_data_watch(
            name=watch_name(fn),
            watch=DelayedWatch(fn, *args, **kwargs),
        )
        return fn

    return decorator


def watch_name(fn):
    return fn.__name__.lstrip("gen_")



import kopf

import digi.util as util
from digi.mount import Mounter


def run():
    import digi

    # mounter
    if digi.enable_mounter:
        Mounter(digi.g, digi.v, digi.r, digi.n, digi.ns,
                log_level=digi.log_level).start()

    digi.model = digi.control.create_model()
    digi.pool = digi.data.create_pool()
    digi.router = digi.data.create_router()

    # reconciler
    _model = {
        "group": digi.g,
        "version": digi.v,
        "plural": digi.r,
    }
    _registry = util.KopfRegistry()
    _kwargs = {
        "when": lambda name, namespace, **_: name == digi.n and namespace == digi.ns,
        "registry": _registry,
    }
    _ready, _stop = None, None

    # force decorate the handlers
    from . import handler
    from .data import router as sync_handler
    _, _ = handler, sync_handler

    @kopf.on.startup(registry=_registry)
    def configure(settings: kopf.OperatorSettings, **_):
        settings.persistence.progress_storage = kopf.AnnotationsProgressStorage()
        settings.posting.level = digi.log_level

    # reconciler operations
    from digi.reconcile import rc
    from digi.view import CleanView
    @kopf.on.create(**_model, **_kwargs)
    @kopf.on.resume(**_model, **_kwargs)
    @kopf.on.update(**_model, **_kwargs)
    def reconcile(spec, meta, *args, **kwargs):
        digi.logger.debug(f"processing gen: {meta['generation']}; "
                          f"resource version: {meta['resourceVersion']}; "
                          f"skip_gen: {rc.skip_gen}; "
                          f"last_seen_gen: {rc.last_seen_gen}; "
                          f"reconcile count: {rc.count}")

        gen = meta["generation"]
        if gen == rc.last_seen_gen:
            digi.logger.info(f"skipped gen {gen} due to last-seen")
            return

        rc.last_seen_gen = gen
        rc.count += 1

        if digi.pool is not None:
            try:
                model = CleanView(dict(spec),
                                  trim_mount=digi.load_trim_mount).m()
                digi.pool.load([model], branch="model")
                digi.logger.info(f"done loading model snapshot to pool")
            except Exception as e:
                digi.logger.warning(f"unable to load to pool: {e}")

        # skip the last self-write
        # TBD for parallel reconciliation may need to lock rc.gen before patch
        if gen == rc.skip_gen and rc.should_skip():
            digi.logger.info(f"skipped gen {gen} due to self-write")
            return

        spec = rc.run(spec, *args, **kwargs)
        _, resp, e = util.check_gen_and_patch_spec(digi.g, digi.v, digi.r, digi.n, digi.ns,
                                                   spec, gen=gen)
        if e is not None:
            if e.status == util.DriverError.GEN_OUTDATED:
                digi.logger.warning(f"gen {gen} outdated; "
                                    f"pending {len(rc._pending_handler)} handlers")
                return
            else:
                raise kopf.PermanentError(e.status)

        # if the model isn't updated do not
        # increment the skip marker
        new_gen = resp["metadata"]["generation"]
        if gen + 1 == new_gen:
            rc.skip_gen = new_gen

        rc.clear_pending()
        digi.logger.info(f"done reconciliation")

    @kopf.on.delete(**_model, **_kwargs, optional=True)
    def on_delete(*args, **kwargs):
        _, _ = args, kwargs
        _stop.set()

    if digi.pool is not None:
        digi.pool.create_branch_if_not_exist("model")

    if digi.enable_visual:
        import os, sys, subprocess
        try:
            subprocess.check_call(f"python visual.py &",
                                  env={**os.environ.copy(), **{
                                      "MOUNTER": "false",
                                      "VISUAL": "false",
                                      "LOGGER_NAME": "digi.visual",
                                  }},
                                  shell=True)
            digi.logger.info("started visualizer")
        except subprocess.CalledProcessError:
            digi.logger.fatal("unable to start visualizer")
            sys.exit(1)

    # TBD graceful termination
    _ready, _stop = util.run_operator(_registry, log_level=digi.log_level)
    digi.rc.start_data_watch()



import os
import sys
import logging
from collections import defaultdict
import kopf

import digi.util as util
from digi.util import parse_gvr, spaced_name, parse_spaced_name

"""
An embedded meta-actor that implements the mount semantics.

Event watch:
- Watch the parent model and identify the child models;
- Watch the child models;

Event propagation:
- On child's updates (intent and status): propagate to the child's 
  copy in the parent;
- On parent's updates (intent) on the child's copy: propagate to 
  the child's intent;
"""

_mount_mode = os.environ.get("MOUNT_MODE", "permissive")
TRIM_FROM_PARENT, TRIM_FROM_CHILD = set(), set() # don't trim any
if _mount_mode == "strict":
    TRIM_FROM_PARENT = {"status", "output", "obs", "meta"}
    TRIM_FROM_CHILD  = {"intent", "input"}
elif _mount_mode == "intent_rec":
    TRIM_FROM_CHILD = set()

class Watch:
    def __init__(self, g, v, r, n, ns="default", *,
                 create_fn=None,
                 resume_fn=None,
                 update_fn=None,
                 # TBD enable finalizer but avoid looping with multiple children
                 delete_fn=None, delete_optional=True,
                 field_fn=None, field="",
                 log_level=logging.INFO):
        self._registry = util.KopfRegistry()
        self._log_level = log_level
        _args = (g, v, r)
        _kwargs = {
            "registry": self._registry,
            # watch a specific model only
            "when": lambda name, namespace, **_: name == n and namespace == ns,
        }

        @kopf.on.startup(registry=self._registry)
        def configure(settings: kopf.OperatorSettings, **_):
            settings.persistence.progress_storage = kopf.AnnotationsProgressStorage()
            settings.posting.level = log_level

        if create_fn is not None:
            kopf.on.create(*_args, **_kwargs)(create_fn)
        if resume_fn is not None:
            kopf.on.resume(*_args, **_kwargs)(resume_fn)
        if update_fn is not None:
            kopf.on.update(*_args, **_kwargs)(update_fn)
        if delete_fn is not None:
            kopf.on.delete(*_args, **_kwargs, optional=delete_optional)(delete_fn)
        if field_fn is not None and field != "":
            kopf.on.field(field=field, *_args, **_kwargs)(field_fn)
        assert create_fn or resume_fn or update_fn or delete_fn, "no handler provided"

        self._ready_flag, self._stop_flag = None, None

    def start(self):
        self._ready_flag, self._stop_flag = util.run_operator(
            self._registry, log_level=self._log_level,
            skip_log_setup=True,
        )
        return self

    def stop(self):
        assert self._stop_flag, "watch has not started"
        self._stop_flag.set()
        return self


class Mounter:
    """Implements the mount semantics for a given (parent) digivice"""

    def __init__(self, g, v, r, n, ns="default",
                 log_level=logging.INFO):

        """ children event handlers """

        def on_child_create(body, meta, name, *args, **kwargs):
            _g, _v, _r = util.gvr_from_body(body)
            self._logger.info(f"on create child {name} gen {meta['generation']}")
            _sync_from_parent(_g, _v, _r, meta=meta, name=name,
                              attrs_to_trim=TRIM_FROM_PARENT,
                              *args, **kwargs)
            _sync_to_parent(_g, _v, _r, meta=meta, name=name,
                            attrs_to_trim=TRIM_FROM_CHILD,
                            *args, **kwargs)

        def on_child_update(body, meta, name, namespace,
                            *args, **kwargs):
            _g, _v, _r = util.gvr_from_body(body)
            _id = util.model_id(_g, _v, _r, name, namespace)

            self._logger.info(f"on child {name} gen {meta['generation']}")
            if meta["generation"] == self._children_skip_gen.get(_id, -1):
                self._logger.info(f"skipped child {name} gen {meta['generation']}")
                return

            return _sync_to_parent(_g, _v, _r, name, namespace, meta,
                                   *args, **kwargs)

        def on_child_delete(body, name, namespace,
                            *args, **kwargs):
            _, _ = args, kwargs

            _g, _v, _r = util.gvr_from_body(body)

            # remove watch
            gvr_str = util.gvr(_g, _v, _r)
            nsn_str = util.spaced_name(name, namespace)

            w = self._children_watches.get(gvr_str, {}).get(nsn_str, None)
            if w is not None:
                w.stop()
                self._children_watches[gvr_str].pop(nsn_str, "")

            # will delete from parent
            _sync_to_parent(_g, _v, _r, name, namespace, spec=None,
                            *args, **kwargs)

        def _sync_from_parent(group, version, plural, name, namespace, meta,
                              attrs_to_trim=None, *args, **kwargs):
            _, _ = args, kwargs

            parent, prv, pgn = util.get_spec(g, v, r, n, ns)

            # check if child exists
            mounts = parent.get("mount", {})
            gvr_str = util.gvr(group, version, plural)
            nsn_str = util.spaced_name(name, namespace)

            if (gvr_str not in mounts or
                    (nsn_str not in mounts[gvr_str] and
                     name not in mounts[gvr_str])):
                self._logger.warning(f"unable to find the {nsn_str} or {name} in the {parent}")
                return

            models = mounts[gvr_str]
            n_ = name if name in models else nsn_str

            patch = models[n_]
            if attrs_to_trim is not None:
                patch = util.trim_attr(patch, attrs_to_trim)

            _, resp, e = util.check_gen_and_patch_spec(
                group, version, plural,
                name, namespace,
                patch, gen=meta["generation"]
            )

            if e is not None:
                self._logger.warning(f"unable to sync from parent to {name} due to {e}")
            else:
                model_id = util.model_id(group, version, plural,
                                         name, namespace)
                new_gen = resp["metadata"]["generation"]
                self._children_gen[model_id] = new_gen
                if meta["generation"] + 1 == new_gen:
                    self._children_skip_gen[model_id] = new_gen

        def _sync_to_parent(group, version, plural, name, namespace, meta,
                            spec, diff, attrs_to_trim=None, *args, **kwargs):
            _, _ = args, kwargs

            # propagation from child retries until succeed
            while True:
                parent, prv, pgn = util.get_spec(g, v, r, n, ns)

                # check if child exists
                mounts = parent.get("mount", {})
                gvr_str = util.gvr(group, version, plural)
                nsn_str = util.spaced_name(name, namespace)

                if (gvr_str not in mounts or
                        (nsn_str not in mounts[gvr_str] and
                         name not in mounts[gvr_str])):
                    self._logger.warning(f"unable to find the {nsn_str} or {name} in the {parent}")
                    return

                models = mounts[gvr_str]
                n_ = name if name in models else nsn_str

                if spec is None:
                    parent_patch = None  # will convert to json null
                else:
                    if models[n_].get("mode", "hide") == "hide":
                        if attrs_to_trim is None:
                            attrs_to_trim = set()
                        attrs_to_trim.add("mount")

                    # TBD rename to _gen_parent_spec
                    parent_patch = _gen_parent_patch(spec, diff, attrs_to_trim)

                # add roots
                if parent_patch is None:
                    # only child
                    if len(mounts[gvr_str]) == 1:
                        parent_patch = {
                            "mount": {
                                gvr_str: None
                            }
                        }
                    else:
                        parent_patch = {
                            "mount": {
                                gvr_str: {
                                    n_: None
                                }
                            }
                        }
                else:
                    parent_patch = {
                        "mount": {
                            gvr_str: {
                                n_: {
                                    "spec": parent_patch,
                                    "generation": meta["generation"],
                                }
                            }
                        }}

                # maybe rejected if parent has been updated;
                # continue to try until succeed
                resp, e = util.patch_spec(g, v, r, n, ns, parent_patch, rv=prv)
                if e is not None:
                    if e.status == 409:
                        self._logger.warning(f"unable to sync to parent from {name} due to conflict; retry")
                    else:
                        self._logger.error(f"failed to sync {name} to parent due to {e}; abort")
                        return
                else:
                    new_gen = resp["metadata"]["generation"]
                    self._logger.info(f"update child {name} generation to {meta['generation']}")
                    if pgn + 1 == new_gen:
                        self._parent_skip_gen = new_gen
                    break

        def _gen_parent_patch(child_spec, diff, attrs_to_trim=None):
            child_spec = dict(child_spec)

            if diff is not None:
                child_spec = util.apply_diff({"spec": child_spec}, diff)["spec"]

            if attrs_to_trim is not None:
                child_spec = util.trim_attr(child_spec, attrs_to_trim)

            return child_spec

        """ parent event handlers """

        def on_parent_create(spec, diff, *args, **kwargs):
            _, _ = args, kwargs
            _update_children_watches(spec.get("mount", {}))
            _sync_to_children(spec, diff)

        def on_mount_attr_update(spec, meta, diff, *args, **kwargs):
            _, _ = args, kwargs

            if meta["generation"] == self._parent_skip_gen:
                self._logger.info(f"skipped parent gen {self._parent_skip_gen}")
                return

            mounts = spec.get("mount", {})

            _update_children_watches(mounts)
            _sync_to_children(spec, diff)
            _prune_mounts(mounts, meta)

        def on_parent_delete(*args, **kwargs):
            _, _ = args, kwargs
            self.stop()

        def _prune_mounts(mounts, meta):
            rv = meta["resourceVersion"]
            while True:
                to_prune = list()
                for gvr_str, models in mounts.items():
                    if len(models) == 0:
                        to_prune.append(gvr_str)
                if len(to_prune) == 0:
                    return
                patch = {
                    "mounts": {
                        gvr_str: None for gvr_str in to_prune
                    }
                }
                _, e = util.patch_spec(g, v, r, n, ns, patch, rv=rv)
                if e is None:
                    self._logger.info(f"prune mount: {patch}")
                    return
                elif e.status != 409:
                    self._logger.warning(f"prune mount failed due to {e}")
                    return

                self._logger.info(f"prune mount will retry due to: {e}")
                spec, rv, _ = util.get_spec(g, v, r, n, ns)
                mounts = spec.get("mount", {})

        def _update_children_watches(mounts: dict):
            # iterate over mounts and add/trim child event watches
            # add watches
            for gvr_str, models in mounts.items():
                gvr = parse_gvr(gvr_str)  # child's gvr

                for nsn_str, m in models.items():
                    nsn = parse_spaced_name(nsn_str)
                    # in case default ns is omitted in the model
                    nsn_str = spaced_name(*nsn)

                    if gvr_str in self._children_watches and \
                            nsn_str in self._children_watches[gvr_str]:
                        continue

                    # TBD: add child event handlers
                    self._logger.info(f"new watch for child {nsn_str}")
                    self._children_watches[gvr_str][nsn_str] \
                        = Watch(*gvr, *nsn,
                                create_fn=on_child_create,
                                resume_fn=on_child_create,
                                update_fn=on_child_update,
                                delete_fn=on_child_delete,
                                log_level=log_level).start()

            # trim watches no longer needed
            for gvr_str, model_watches in self._children_watches.items():
                mw_to_delete = set()
                for nsn_str, w in model_watches.items():
                    models = mounts.get(gvr_str, {})
                    if nsn_str not in models and \
                            util.trim_default_space(nsn_str) not in models:
                        w.stop()
                        mw_to_delete.add(nsn_str)

                for d in mw_to_delete:
                    model_watches.pop(d, None)

        def _gen_child_patch(parent_spec, gvr_str, nsn_str):
            mount_entry = parent_spec \
                .get("mount", {}) \
                .get(gvr_str, {}) \
                .get(nsn_str, {})
            if mount_entry.get("mode", "hide") == "hide":
                mount_entry.get("spec", {}).pop("mount", {})

            if mount_entry.get("status", "inactive") == "active":
                spec = mount_entry.get("spec", None)
                if spec is not None:
                    spec = util.trim_attr(spec, TRIM_FROM_PARENT)

                gen = mount_entry.get("generation", sys.maxsize)
                return spec, gen

            return None, None

        def _sync_to_children(parent_spec, diff):
            # sort the diff by the attribute path (in tuple)
            diff = sorted(diff, key=lambda x: x[1])

            # filter to only the intent/input updates
            to_sync = dict()
            for _, f, _, _ in diff:
                # skip non children update
                if len(f) < 3:
                    continue

                gvr_str, nsn_str = f[0], f[1]
                model_id = util.model_id(*parse_gvr(gvr_str),
                                         *parse_spaced_name(nsn_str))

                if model_id not in to_sync:
                    cs, gen = _gen_child_patch(parent_spec, gvr_str, nsn_str)
                    if cs is not None:
                        to_sync[model_id] = cs, gen

            # sync all, e.g., on parent resume and creation
            if len(diff) == 0:
                for gvr_str, ms in parent_spec.get("mount", {}).items():
                    for nsn_str, m in ms.items():
                        model_id = util.model_id(*parse_gvr(gvr_str),
                                                 *parse_spaced_name(nsn_str))
                        cs, gen = _gen_child_patch(parent_spec, gvr_str, nsn_str)
                        # both rv and gen can be none as during the initial sync
                        # the parent may overwrite
                        if cs is not None:
                            to_sync[model_id] = cs, gen

            # push to children models
            # TBD: transactional update
            for model_id, (cs, gen) in to_sync.items():
                cur_gen, resp, e = util.check_gen_and_patch_spec(
                    *util.parse_model_id(model_id),
                    spec=cs,
                    gen=max(gen, self._children_gen.get(model_id, -1)))
                if e is not None:
                    self._logger.warning(f"unable to sync to child {model_id} due to {e}")
                else:
                    new_gen = resp["metadata"]["generation"]
                    self._children_gen[model_id] = new_gen
                    if cur_gen + 1 == new_gen:
                        self._children_skip_gen[model_id] = new_gen

        # subscribe to the events of the parent model
        self._parent_watch = Watch(g, v, r, n, ns,
                                   create_fn=on_parent_create,
                                   resume_fn=on_parent_create,
                                   field_fn=on_mount_attr_update, field="spec.mount",
                                   delete_fn=on_parent_delete, delete_optional=True,
                                   log_level=log_level)

        # subscribe to the events of the child models;
        # keyed by the gvr and then spaced name
        self._children_watches = defaultdict(dict)

        # last handled generation of a child, keyed by model_id;
        # used when update the children because the parent's copy
        # might be out of date
        self._children_gen = dict()

        # used to filter last self-write on the child
        self._children_skip_gen = dict()
        self._parent_skip_gen = -1

        # mounter logging
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(log_level)

    def start(self):
        self._parent_watch.start()
        self._logger.info("started the mounter")

    def stop(self):
        self._parent_watch.stop()
        for _, mws in self._children_watches.items():
            for _, w in mws.items():
                w.stop()
        return self


def test():
    gvr = ("mock.digi.dev", "v1", "samples")
    Mounter(*gvr, n="sample").start()


if __name__ == '__main__':
    test()



import sys
import digi


class Model():
    def get(self):
        return digi.rc.view()

    def patch(self, view_or_path, value=None, gen=sys.maxsize):
        """If first arg given is a view patch it directly; if not
        use it as the path and use the second to compose the view.
        This is useful when the updated value has a prefix."""
        if isinstance(view_or_path, dict):
            view = view_or_path
        elif isinstance(view_or_path, str):
            path, view = view_or_path, dict()
            fields = path.strip(".").split(".")
            front = view
            for field in fields[:-1]:
                front[field] = dict()
                front = front[field]
            front[fields[-1]] = value
        else:
            raise NotImplementedError

        cur_gen, r, e = digi.util.check_gen_and_patch_spec(
            digi.g, digi.v, digi.r, digi.n, digi.ns,
            view, gen)

        if e != None:
            digi.logger.info(f"patch error: {e}")
        return cur_gen, r, e

    def update(self, *args, **kwargs):
        return self.patch(*args, **kwargs)

    def get_mount(self,
                  group=digi.name,
                  version=digi.version,
                  resource=None,
                  any: bool = False,
                  ) -> dict:
        """If any is set, returns all mounts as name:mount pairs.
        If resource is given, returns all mounts under the resource.
        Otherwise returns the mount root."""
        if any:
            mounts = dict()
            for _, name_mount in digi.rc.view().get("mount", {}).items():
                for name, mount in name_mount.items():
                    mounts[name] = mount
            return mounts
        if resource:
            path = f"mount.'{group}/{version}/{resource}'"
        else:
            path = "mount"
        return digi.util.get(digi.rc.view(), path)


def create_model():
    return Model()



import time
import yaml
import os
import sys
import pprint as pp

_dir = os.path.dirname(os.path.realpath(__file__))
_parent_dir = os.path.dirname(_dir)
sys.path.insert(0, _parent_dir)

from digi.processor import jq

test_yaml = f"""
control:
  brightness:
    intent: 0.8
    status: 0
  mode:
    intent: sleep
    status: sleep
mount:
  mock.digi.dev/v1/lamps:
    default/lamp-test:
      spec:
        control:
          power: 
            intent: "on"
  mock.digi.dev/v1/motionsensors:
    default/motionsensor-test:
      spec:
        obs:
          last_triggered_time: {time.time()}
reflex:
  motion-mode:
    policy: 'if $time - ."motionsensor-test".obs.last_triggered_time <= 600 
             then .root.control.mode.intent = "work" else . end'
    priority: 1
"""

if __name__ == '__main__':
    v = yaml.load(test_yaml, Loader=yaml.FullLoader)
    start = time.time()
    jq(v["reflex"]["motion-mode"]["policy"])(v)
    print(f"took {time.time() - start}s")
    pp.pprint(v)



#!/usr/bin/env python3

import digi

import dash
from dash import dcc
from dash import html
from dash import dash_table
from dash.dependencies import Input, Output, State

import pandas as pd

app = dash.Dash(digi.name)
elements = [html.Button(digi.name)]

app.layout = html.Div([
    html.Div(id="digi"),
    html.Div(id="data"),
    dcc.Interval(
        id='interval',
        interval=digi.visual_refresh_interval,
        n_intervals=0
    ),
])


@app.callback(
    Output('digi', 'children'),
    Output('data', 'children'),
    Input('interval', 'n_intervals'),
)
def update_layout(n):
    digis, models = list(), digi.view.NameView(digi.rc.view()).m()
    digis.append(html.Button(digi.name))
    for name, _ in models.items():
        if name == "root":
            continue
        digis.append(html.Button(name))

    readings = digi.pool.query(f"from {digi.name} | not _type=='model' | head 10")
    df = pd.DataFrame(readings)
    data = dash_table.DataTable(
        id='table',
        columns=[{"name": i, "id": i} for i in df.columns],
        data=df.to_dict('records'),
    )
    return digis, data


def main():
    digi.run()
    app.run_server(debug=True, port=7534)


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

URLS = {
}

setup(
    name="digi",
    version="0.2.5",
    description="Digi driver library",
    url=URLS,
    author="Team Digi",
    author_email="silveryfu@gmail.com",
    license="BSD-3-Clause License",
    packages=find_packages(exclude=("tests",)),
    python_requires='>=3.7',
    include_package_data=True,
    install_requires=[
        "pyyaml",
        "inflection",
        "python-box",
        "kubernetes",
        'kopf',
        "zed",
        "dash",
        "pandas",
    ],
)
